// --- [ main.c ] -------------------------------------------------------------

/*

Messung von Akkukapazitten mit automatisch Abschaltung bei festgelegter
Mindestspannung.

Die Zeitbasis ist 1 Sekunde.
Sie wird vom Timer1 geliefert, der von 0 bis TCR1A luft,
dabei automatisch den TNCT1 zurcksetzt und einen CompareA-Interrupt aufruft.
Dabei wird ein Sekundenflag gesetzt, das in main() ausgewertet wird.

Dort werden zwei Trigger decrementiert.
Erreichen sie Null, dann werden Aktionen ausgelst:
Trigger1 -> fhrt eine Messung aus
Trigger2 -> schreibt den aktuellen Messwerte ins TWI-Eeprom raus.

Die Kommunikation von und nach aussen erfolgt ber die USI-TWI-Schnittstelle.
Von aussen knnen:
- Beim Programmstart knnen im Eeprom des Tiny gespeicherte Einstellungen ausgelesen werden.
- Whrend der Messung die aktuellen Messwerte gelesen werden.
- Im Programmcode voreingestellte Wert fr die TWI-Adresse, Referenezspannung,
  Entladeschlussspannung und Widerstand gendert werden.

Gleichzeitig kann der Tiny gemessene bzw. berechnete Daten via USI-TWI-Master auf ein externes TWI-Device
geschrieben werden.
Das knnte ein TWI-Display sein.
Oder wie hier ein TWI-Eeprom, auf dem die Daten zur spteren Auswertung gespeichert werden.

Die Messung erfolgt alternativ nach 2 Methoden:

1.) REFERENZ = R_EXTERN
Eine externe Referenzspannung wird an PA0 angelegt (oder es wird VCC als Referenz verwendet).
Die zu messende Akkuspannung wird an ADC7 angeschlossen.

2.) REFERENZ = R_INTERN
Die interne 1.1V Referenz wird genutzt.
In diesem Falle wird die zu messende Spannung an PA0 erwartet.
Aus der Abweichung der gemessenen internen Referenz vom Sollwert kann die Spannung an PA0 errechnet werden.

Der Entladestrom fliet ber einen konstanten Widerstand R.
Mit der Formel I = U / R kann der flieende Strom berechnet werden.
Die Messwerte werden im 6 Sekunden-Abstand aufaddiert, so dass am Ende der Gesamtbetrag
durch 600 dividiert werden muss, um die geflossen Ladung in mAh ermitteln zu knnen.

Auf dem ATtiny24 kann nicht die gesamte Funktionalitt untergebracht werden.
Daher wird die Berechnung des Stroms sowie das Aufaddieren der Messwerte dem
auswertenden Programm auf dem PC berlassen.

*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <inttypes.h>
#include <stdlib.h>

#include "global.h"
#include <util/delay.h>
#include "USI_Slave_e.h"
#include "USI_Master.h"
#include "eeprom.h"

// --- Initialisierung
// - GPIO-Register fr die volatile-Variablen 'new_tick' und 'state' nutzen
// - die Assembler ISR-Routinen benutzen diese Register !!!!
//volatile uint8_t new_tick;
#define new_tick     GPIOR0
//volatile uint8_t state;
#define state        GPIOR1

uint32_t sum_i;				   // Kapazitt * (3600 / INTERVALL_1)
uint16_t sum_j;               // sum_i / (3600 / INTERVALL_1)
uint16_t sum_t;               // Summe der Einzelmessungen
uint16_t ist_u, ist_i;			// letzter Messwert fr U und I
uint16_t eeprom_adr;          // TWI-Adresse des TWI-Eeproms
uint16_t u_ref;               // Referenzspannung und Widerstand als Variable
uint16_t r_ref;               // knnen via USI gendert werden
uint16_t u_fin;               // Entladeschlussspannung
uint8_t  out[20];             // Array zur Ausgabe ins USI-Modul

uint8_t twi_err = FALSE;
uint8_t id;

// ---------------------------------------------------------------------------
uint16_t messen(void);
void write_array(uint8_t idx);
void end_discharge(void);
void usi_twi_send(uint8_t cnt);
void daten_schreiben(void);


// --- Main -------------------------------------------------------------------

int main(void)
	{
   uint32_t v;
	uint8_t trigger1 = 0, trigger2 = 0, i, j;
	uint8_t ende = FALSE;

	CLKPR = (1<<CLKPCE);
	CLKPR = (0<<CLKPS3 | 0<<CLKPS2 | 0<<CLKPS1 | 1<<CLKPS0); // Prescale 2 -> 4.0MHz

//	PRR = (0<<PRTIM1 | 1<<PRTIM0 | 0<<PRUSI | 0<<PRADC);  // berflssiges Abschalten

	// --- Pins fr LED's und FET setzen
	DDRA  = (1<<FET | 1<<LED_YEL | 1<<LED_GRN | 1<<LED_ERR);
	PORTA = (1<<LED_YEL);            // LED einschalten,
	PORTB = (1<<TASTE);              //Pullup fr Taste setzen

   // --- Referenzspannung einlesen - falls etwas im EEPROM steht
   // tt = (EE_Read(5) << 8) + EE_Read(6);
   // if (tt < 0xFFFF) u_ref = tt;
   i = EE_Read(2);
   j = EE_Read(3);
   if (!((j == 255) && (j == 255))) u_ref = (i<<8) + j;
#if REFERENZ == R_INTERN
   else u_ref = VREFi;
#elif REFERENZ == R_EXTERN
   else u_ref = VREFe;
#else
   #error Keine Referenz angegeben !!!!
#endif

   // --- Entladeschlussspannung einstellen
   i = EE_Read(4);
   j = EE_Read(5);
   if (!((j == 255) && (j == 255))) u_fin = (i<<8) + j;
   else u_fin = VFIN;

#if READ_R == TRUE
   // --- Widerstand einlesen - falls etwas im EEPROM steht
   //tt = (EE_Read(3) << 8) + EE_Read(4);
   //if (tt < 0xFFFF) r_ref = tt;
   i = EE_Read(6);
   j = EE_Read(7);
   if (!((j == 255) && (j == 255))) r_ref = (i<<8) + j;
   else r_ref = R;

#endif

   asm volatile("sei");

   // ds erste Byte aus dem TWI-EEprom auslesen,
   // es wird spter incrementiert und als erstes Byte jedem
   // Datensatz vorangestellt,
   // Eine Messreihe wird durch Datenstze mit gleiche ID markiert.

   USI_MA_Init();
   out[0] = TWI_EEPROM_ADR;
   out[1] = 0;
   out[1] = 0;
   USI_MA_Write(out, 3);
   USI_MA_Read(out, 2);
   id = out[1];                  // Das erste Byte des TWI-Eproms

   USI_SLA_Init();

   // die ersten 8 Byte des Tiny-Eeproms ins USI-Modul exportieren
   // zur Kontrolle der dort abgelegten Einstellungen
   for(i = 0; i<8; i++)
      {
      out[i] = EE_Read(i);
      }
   USI_SLA_Put_Data(out, 8);

   messen();               // erste (fehlerbehaftete) Messung ausfhren und das Ergebnis ignorieren
   state = 0;

	while (TRUE)
		{
		// in den Wartezustand gehen, bis die Taste an PB2 bettigt wird
		// und einen PinChangeInterrupt auslst - oder eine TWI-Transaktion stattfindet
		if (state == 0)
			{
				TCCR1B = TIMER1_OFF;                // Timer1 abschalten
				sum_i  = 0;
				sum_t  = 0;
				eeprom_adr = 0;
				id++;                               // Die ID bei jeder neuen Messreihe incrementieren
				trigger1 = 1;                       // damit sofort eine Messung ausgefhrt wird
				trigger2 = 1;                       // um die Messung zu schreiben

				twi_err = FALSE;

				MCUCR  = (1<<SE | 1<<SM1 | 0<<SM0); // Powerdown-Modus

				GIFR   = (1<<PCIF1);                // falls noch ein Flag gesetzt ist
				PCMSK1 = (1<<TASTE_INT);            // PCINT10 an PB2
 				GIMSK  = (1<<PCIE1);

				// ---------------------------------------------------------------------
				asm volatile("sleep");
				// ---------------------------------------------------------------------

				GIMSK = 0;
				MCUCR = 0;

            // Dann fand eine USI-Transaktion statt --------------------------------
				if (state == 0)
               {
               i = USI_SLA_Rx_Cnt();            // sind Daten empfangen worden ?
               if (i == 4)                      // wenn genau 4 Byte
                  {
                  PINA = (1<<LED_ERR);          // Rote LED toggle
                  USI_SLA_Get_Data(out);        // empfangene Daten in out[] einlesen

                  if (out[1] == 100)            // 1. Byte == 100
                     {
                     if (out[2] == 200)         // Entladung starten
                        {
                        state = 1;
                        }
                     }

                  else if (out[1] == 251)        // Adressnderung
                     {
                     USI_Write_TWI_Adr(&out[2]);
                     }

                  else if (out[1] == 252)        // Referenzspannung
                     {
                     EE_Write(2, out[2]);
                     EE_Write(3, out[3]);
                     }

                  else if (out[1] == 253)       // Entladeschlussspannung
                     {
                     EE_Write(4, out[2]);
                     EE_Write(5, out[3]);
                     }

#if READ_R == TRUE
                  else if (out[1] == 254)      // Widerstandswert
                     {
                     EE_Write(6, out[2]);
                     EE_Write(7, out[3]);
                     }
#endif
                  }
               if (state == 0) continue;                           // gleich zurck zum Schleifenbeginn
               }

            // Es wurde die Taste gedrckt und die Entladung beginnt ---------

				// Kontroll-LED einschalten (gegen gnd geschaltet)
				// Akku-Masse anschlieen
				PORTA |= (1<<FET | 1<<LED_GRN | 1<<LED_YEL);

				// Timer1 einschalten
				TCNT1H = 0;
				TCNT1L = 0;

				// Sekundentakt: 4 MHz 4.000.000 / 256 = 15.625
				// F = FCPU / (2 * 256 *(1 + OCR1A) = 4.000.000 / (2 * 256 * 15.625) = 0.5 Hz
				OCR1AH = (uint8_t) (15624 >> 8);
				OCR1AL = (uint8_t) 15624;
				// CTC-Mode, Prescale 256
				TCCR1B = TIMER1_ON;
				TIMSK1 = (1<<OCIE1A);
			}

		while (state)
			{
         i = USI_SLA_Rx_Cnt();            // sind Daten via TWI empfangen worden ?
         if (i == 2)                      // wenn genau 2 Byte
            {
            USI_SLA_Get_Data(out);        // empfangene Daten in out[] einlesen
            if (out[1] == 201)            // Entladung unterbrechen
               {
               PORTA &= ~(1<<FET | 1<< LED_GRN);
               PORTA |=  (1<<LED_YEL | 1<<LED_ERR);
               TCCR1B = TIMER1_OFF;
               }

            else if (out[1] == 202)      // Entladung fortsetzen
               {
               PORTA |= (1<<FET | 1<<LED_GRN);
               PORTA &=  ~(1<<LED_YEL | 1<<LED_ERR);
               TCCR1B = TIMER1_ON;
               }

            else if (out[1] == 203)       // Entladung beenden
               {
               ende = TRUE;
               }
            }

			if(new_tick)                  // wenn wieder 1 Sekunde um ist
				{
				new_tick = FALSE;

				PINA = (1<<LED_YEL);			// LED schaltet im Sekundentakt

            if (--trigger1)            // Ist der Zeitpunkt fr eine Messung gekommen ?
               {
               continue;               // wenn nicht, dann zurck zu while
               }

            else                       // trigger wurde nach 0 decrementiert
               {

               trigger1 = INTERVALL_1;
               trigger2--;
               }

				ist_u = messen();          // Spannungsmessung ausfhren

				// Messwert ber Vref auf Spannung hochrechnen
				// v ist 32Bit breit

#if REFERENZ == R_INTERN
            v = ((uint32_t) u_ref * 1023) / ist_u;
//          v = ((uint32_t) VREFe) * 1023) / ist_u;
#elif REFERENZ == R_EXTERN
      		v = ((uint32_t) u_ref * ist_u) / 1023;
//          v = ((uint32_t) VREFe * ist_u) / 1023;
#endif
				// nach der Division reduzieren auf 16 Bit
				ist_u = (uint16_t) v;

            write_array(0);            // Daten in das Array out[] ab Pos. 0 schreiben
            USI_SLA_Put_Data(out, 6);  // ins USI-Modul exportieren

            // prfen, ob die Mindestspannung unterschritten ist
				if ((ist_u < u_fin) || ende)
					{
					ende = FALSE;
					// abschlieenden Datensatz schreiben, Entladen beenden
               end_discharge();
					state = 0;  // while(state) beenden
					}

				else if (trigger2 == 0)
               {
               trigger2 = INTERVALL_2;
               daten_schreiben();         // Daten ins Eeprom schreiben
               // dabei kann ein TWI-Fehler auftreten
               // Warnung bei TWI-Fehler ausgeben - LED leuchtet
               if (twi_err) PORTA |= (1<<LED_ERR);
               else PORTA &= ~(1<<LED_ERR);
               }

            // der erste Datensatz wird beim Programmstart geschrieben
            // hier stehen noch alle Zhler auf Null
            // den Zeitzhler erst am Ende incrementieren
				sum_t++;

#if CALC_I == TRUE
            // I = U / R // R ist mit Faktor 10 multipliziert,
            //              daher auch Faktor 10 auf die Spannung
				ist_i = (ist_u * 10) / r_ref;
				sum_i += ist_i;

				// runterrechnen auf mAh
				sum_j = (uint16_t) (sum_i / (3600 / INTERVALL_1));
#endif
				}
			}
		}
	}

// ----------------------------------------------------------------------------
// schreibt die auszugebenden Werte ins Array out[] ab Pos idx

void write_array(uint8_t idx)
	{
	// Zeitdauer in der Einheit INTERVALL_1 (6 Sekunden)
   out[idx++] = (uint8_t) (sum_t >> 8);
   out[idx++] = (uint8_t)  sum_t;

   // bisherige 'Entladekapazitt'
   out[idx++] = (uint8_t) (sum_j >> 8);
   out[idx++] = (uint8_t)  sum_j;

   // der letzte gemessen Spannungswert
   out[idx++] = (uint8_t) (ist_u >> 8);
   out[idx]   = (uint8_t)  ist_u;
	}

// ----------------------------------------------------------------------------
// Schreibt die auszuwertenden Daten zunchst ins Array out[]
// und anschlieend auf ein TWI-EEPROM
void daten_schreiben(void)
{
   out[0] = TWI_EEPROM_ADR;  // Header mit Adressen
   out[1] = (uint8_t) (eeprom_adr >> 8);
   out[2] = (uint8_t) eeprom_adr;
   out[3] = id;

   eeprom_adr += 8;

   write_array(4);   // Daten ergnzen

   // das letzte Byte auffllen
   out[10] = 255;

   usi_twi_send(11);
}


// ----------------------------------------------------------------------------
// sendet als USI-Master vorbereitete Daten im Array out[]
void usi_twi_send(uint8_t cnt)
{
   USI_MA_Init();
   if( USI_MA_Write(out, cnt) ) twi_err = FALSE;
   else twi_err = USI_MA_Error_State();
   USI_SLA_Init();
}


// -----------------------------------------------------------------------------
// Entladung beenden ,
// letzten Datenstand schreiben,
// abschlieenden Datensatz mit 4x 0xFF sowie den Werten
// fr den Entladewiderstand und die Referenzspannung ausgeben.
void end_discharge(void)
   {
   // Akku abschalten, LED auf Dauerlicht
	PORTA &= ~(1<<FET | 1<<LED_GRN);
	PORTA |=  (1<<LED_YEL);

   daten_schreiben();  // letzten Stand schreiben,dabei eeprom_adr incrementieren

	//out[0] = TWI_EEPROM_ADR;            // die Adr ist schon eingetragen
   out[1] = (uint8_t) (eeprom_adr >> 8);
   out[2] = (uint8_t) eeprom_adr;

   // in out[3] steht noch die id
   // Die nchsten 3 Byte des letzten Datensatzes bestehen aus 0XFF
   // dann  folgen die Werte fr U-Ref und R-Ref

   out[4] = (uint8_t) (u_ref >> 8);
   out[5] = (uint8_t) u_ref;

   out[6] = (uint8_t) (u_fin >> 8);
   out[7] = (uint8_t) u_fin;

   out[8] = (uint8_t) (r_ref >> 8);
   out[9] = (uint8_t) r_ref;
   out[10] = 0;                  // "0" als Zeichen fr den letzten Datensatz
                                 // alle anderen Datenstze enden mit 0xFF

   _delay_ms(EEPROM_WAIT_TIME); // Erholungszeit nach dem letzten TWI-Write
   usi_twi_send(11);
   }

// ----------------------------------------------------------------------------
// fhrt eine Messungreihe aus 4 Einzelmessungen aus
// und liefert den 16-Bit Messwert als Mittelwert zurck
uint16_t messen(void)
	{
	uint16_t m = 0;
	uint8_t i = 4;

#if REFERENZ == R_EXTERN
	// die externe Referenzspannung ist angelegt an PA0,
	// Messung an PA7
	ADMUX   = (0<<REFS1 | 1<<REFS0 | 0<<MUX5 | 0<<MUX4  | 0<<MUX3 | 1<<MUX2 | 1<<MUX1 | 1<<MUX0);
	ADCSRA  = (1<<ADEN | 1<<ADPS2 | 0<<ADPS1 | 1<<ADPS0);

#elif REFERENZ == R_INTERN
	// die Akkuspannung ist angelegt an PA0,
	// Messung der internen 1.1Volt Referenz
	ADMUX   = (0<<REFS1 | 1<<REFS0 | 1<<MUX5 | 0<<MUX4  | 0<<MUX3 | 0<<MUX2 | 0<<MUX1 | 1<<MUX0);
	ADCSRA  = (1<<ADEN | 1<<ADPS2 | 0<<ADPS1 | 1<<ADPS0);
#endif

   while(i--)
      {
      ADCSRA |= ( 1<<ADSC);
      while (ADCSRA & (1<<ADSC));

      m += ADCL;
      m += 	(ADCH << 8);
      }

	return (m >> 2);
	}


#if IRQ_IN_ASM == FALSE
// aus Platzgrnden in Assembler realisiertbar - spart 36 Bytes ein ...
// ----------------------------------------------------------------------------
// wird jede Sekunde aufgerufen
// schaltet nach x-Sekunden das Flag new_tick auf True
ISR(TIM1_COMPA_vect)
	{
	new_tick = TRUE;
	}


// ----------------------------------------------------------------------------
// wird zum Starten des Entladevorganges genutzt.
// an state kann erkannt werden, ob ein Tastendruck oder
// eine USI-Interrupt das sleep beendet hat.

ISR(PCINT1_vect)
	{
	state = 1;
	}
#endif

// --- [ eof ] ----------------------------------------------------------------
